#!/usr/bin/env python
# file test_profiles.py
# author Florent Guiotte <florent.guiotte@irisa.fr>
# version 0.0
# date 10 déc. 2019

import sap
import pytest
import numpy as np
from matplotlib import pyplot as plt
import tempfile

@pytest.fixture
def image():
    return np.arange(100 * 100).reshape(100, 100)

@pytest.fixture
def profiles(image):
    return sap.attribute_profiles(image, {'area': [10, 100],
                                          'compactness': [.1, .5]},
                                        image_name='image')

@pytest.mark.parametrize('adjacency, attribute, exptd_stacks, exptd_profiles', 
        [(4, {'area': [10, 100]}, 1, (5,)), 
         (8, {'area': [10, 100]}, 1, (5,)),
         (4, {'compactness': [.1, .5], 'volume': [100, 5000, 1000]}, 2, (5, 7))
        ])
def test_attribute_profiles(image, attribute, adjacency,
        exptd_stacks, exptd_profiles):

    aps = sap.attribute_profiles(image, attribute, adjacency)

    assert len(aps) == exptd_stacks, \
    'Expected stacks missmatch'

    for ap, ep in zip(aps, exptd_profiles):
        assert ap.data.shape[0] == ep, 'Expected profiles count missmatch'

def test_attribute_profiles_assert(image):
    with pytest.raises(AttributeError) as e:
        sap.Profiles([image], [{}, {}, {}])

def test_profiles_iter(profiles):
    n = len(profiles)
    i = 0
    for profile in profiles:
        i += 1
    assert n == i, 'Wrong number of profiles expected in iter'


@pytest.mark.parametrize('params', [{}, {'height': 4}, {'fname': 'test.png'}])
def test_show_profiles(profiles, params, tmpdir):
    if 'fname' in params:
        params['fname'] = tmpdir.join(params['fname'])

    sap.show_profiles(profiles[0], **params)
    plt.close()

@pytest.mark.parametrize('params', [{}, {'height': 4}, {'fname': 'test.png'},
    {'image': 'image'}, {'attribute': 'area'}])
def test_show_all_profiles(profiles, params, tmpdir):
    if 'fname' in params:
        params['fname'] = tmpdir.join(params['fname'])

    sap.show_all_profiles(profiles[0], **params)
    plt.close()

def test_differential(profiles):
    sap.differential(profiles)

def test_profiles_m_diff(profiles):
    profiles.diff()



#!/usr/bin/env python
# file __init__.py
# author Florent Guiotte <florent.guiotte@irisa.fr>
# version 0.0
# date 12 nov. 2019
"""
Simple Attribute Profiles
=========================

This package provides:

1. Easy to use tree structures.
2. Straight forward attribute profiles computation.

This package heavily relies on the outstanding library `higra
<https://higra.readthedocs.io>`_ which provide efficient tree structure
in C++. For more specific needs you definitely have to check this out.

Documentation
-------------

Documentation is available on docstrings and on the web page `python-sap
documentation <https://python-sap.rtfd.io>`_.

Once sap imported with:

>>> import sap

Use `help()` on the module, submodules, classes and functions to print
directly the docstrings.

>>> help(sap.trees)

Submodules access
-----------------

For simplicity the submodules classes and function are directly
available at the root of the module. In doing so:

>>> from sap import trees
>>> trees.MaxTree

Is equivalent to:

>>> import sap
>>> sap.MaxTree

"""

from .trees import *
from .profiles import *



#!/usr/bin/env python
# file test_Tree.py
# author Florent Guiotte <florent.guiotte@irisa.fr>
# version 0.0
# date 13 nov. 2019

import pytest
import numpy as np
import sap

@pytest.fixture
def image():
    return np.arange(100 * 100).reshape(100, 100)

@pytest.fixture
def max_tree(image):
    return sap.MaxTree(image)


@pytest.mark.parametrize('adjacency', [4, 8])
def test_MaxTree_constructor(image, adjacency):
    t = sap.MaxTree(image, adjacency)

@pytest.mark.parametrize('adjacency', [4, 8])
def test_MinTree_constructor(image, adjacency):
    t = sap.MinTree(image, adjacency)

def test_available_attributes(max_tree):
    att_dict = max_tree.available_attributes()
    assert len(att_dict) > 0, 'Returned list is empty'

@pytest.mark.parametrize('attribute', [
'area',
'child_number',
'compactness',
'contour_length',
#'contour_strength',
'depth',
'dynamics',
'edge_length',
#'extinction_value',
'extrema',
'frontier_length',
#'frontier_strength',
'gaussian_region_weights_model',
'height',
'lca_map',
'mean_vertex_weights',
#'piecewise_constant_Mumford_Shah_energy',
'regular_altitudes',
'sibling',
'vertex_area',
'vertex_list',
'volume',])
def test_MaxTree_get_attribute(max_tree, attribute):
    attribute = max_tree.get_attribute(attribute)
    assert attribute is not None, 'get_attribute returned nothing'
    assert len(attribute) != 0, 'get_attribute returned an empty list'

def test_num_nodes(max_tree):
    assert max_tree.num_nodes() == 20000, 'Max tree did not returned correct number of nodes.'

def test_reconstruct(max_tree, image):
    deleted_nodes = np.zeros(max_tree.num_nodes(), dtype=np.bool)
    filtered_image = max_tree.reconstruct(deleted_nodes)
    assert (filtered_image == image).all(), 'Did not returned the same input image'

    deleted_nodes = np.ones(max_tree.num_nodes(), dtype=np.bool)
    filtered_image = max_tree.reconstruct(deleted_nodes)
    assert (filtered_image == 0).all(), 'Did not returned the filtered image'

    filtered_image = max_tree.reconstruct()
    assert (filtered_image == image).all(), 'Default input not working'

    filtered_image = max_tree.reconstruct(False)
    assert (filtered_image == image).all(), 'Boolean input not working'



#!/usr/bin/env python
# file test_first.py
# author Florent Guiotte <florent.guiotte@irisa.fr>
# version 0.0
# date 12 nov. 2019

def test_first():
    assert True, 'If it fails there is nothing we can do'



#!/usr/bin/env python
# file profiles.py
# author Florent Guiotte <florent.guiotte@irisa.fr>
# version 0.0
# date 13 déc. 2019
"""
Profiles
========

This submodule contains the attribute profiles related classes.

Example
-------

Create the attribute profiles (AP) of `image` based on area attribute
and three thresholds.

>>> aps = sap.attribute_profiles(image, {'area': [10, 100, 1000]})
>>> aps.vectorize()
[[...]]

Create the extended AP of `image` based on area compactness and volume
attributes.

>>> attributes = {'compactness': [.1, .5, .7], 'volume': [10, 100]}
>>> eaps = sap.attribute_profiles(image, attributes)
>>> eaps.vectorize()
[[...]]

Todo
----

Names:
    - vectorize or concatenate ?

"""

import numpy as np
from tqdm.auto import tqdm
from pprint import pformat
from matplotlib import pyplot as plt
from pathlib import Path
from . import trees

class Profiles:
    """
    Base class for profiles.

    Parameters
    ----------
    data : list of ndarray
        List of ndarray representing profiles grouped by image or attribute
        filtering.
    description : list of dict
        List of dictionary containing the metadata of the profiles.

    """
    def __init__(self, data, description):
        if len(data) != len(description):
            raise AttributeError('Data and description missmatch.')

        self.data = data if len(data) > 1 else data[0]
        self.description = description if len(data) > 1 else description[0]
        self.len = len(data)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return 'Profiles' + pformat(self.description)

    def __iter__(self):
        if self.len == 1:
            yield self
            return

        for data, description in zip(self.data, self.description):
            yield Profiles([data], [description])

    def __len__(self):
        return self.len

    def __getitem__(self, key):
        if self.len == 1 and key == 0:
            return self
         
        return Profiles([self.data[key]], [self.description[key]])
    
    def diff(self):
        """Compute the differential of profiles.

        Returns
        -------
        differential : Profiles
            The processed differential profiles.
        
        """
        return differential(self)


def attribute_profiles(image, attribute, adjacency=4, image_name=None):
    """
    Compute the attribute profiles of an image.

    Parameters
    ----------
    image : ndarray
        The image
    attribute : dict
        Dictionary of attribute (as key, str) with according thresholds
        (as values, number).
    adjacency : int
        Adjacency used for the tree construction. Default is 4.
    image_name : str
        The name of the image (optional). Useful to track filtering
        process and display. If not set, the name is replaced by the
        hash of the image.

    Examples
    --------

    >>> image = np.random.random((100, 100))
    >>> sap.attribute_profiles(image, {'area': [10, 100]})
    Profiles[{'attribute': 'area',
    'image': 6508374204896978831,
    'profiles': [{'operation': 'open', 'threshold': 100},
                 {'operation': 'open', 'threshold': 10},
                 {'operation': 'copy'},
                 {'operation': 'close', 'threshold': 10},
                 {'operation': 'close', 'threshold': 100}]}]

    See Also
    --------
    sap.trees.available_attributes : List available attributes.

    """
    data = []
    description = []

    max_tree = trees.MaxTree(image, adjacency)
    min_tree = trees.MinTree(image, adjacency)

    iter_count = sum(len(x) for x in attribute.values()) * 2 + len(attribute)
    ttq = tqdm(desc='Total', total=iter_count)
    for att, thresholds in attribute.items():
        profiles = []; profiles_description = []
        tq = tqdm(total=len(thresholds) * 2 + 1, desc=att)

        # Open
        prof, desc = _compute_profiles(min_tree, att, thresholds[::-1], 'open', (ttq, tq))
        profiles += prof
        profiles_description += desc

        # Origin
        tq.update(); ttq.update()
        profiles += [image]
        profiles_description += [{'operation': 'copy'}]

        # Close
        prof, desc = _compute_profiles(max_tree, att, thresholds, 'close', (ttq, tq))
        profiles += prof
        profiles_description += desc

        tq.close()

        data += [np.stack(profiles)]
        description += [{'attribute': att, 
                         'profiles': profiles_description,
                         'image': image_name if image_name else
                         hash(image.data.tobytes())}]
    ttq.close()

    return Profiles(data, description)

def _compute_profiles(tree, attribute, thresholds, operation, tqs):
    data = []
    desc = []

    for t in thresholds:
        for tq in tqs: tq.update()
        desc += [{'operation': operation, 'threshold': t}]
        deleted_nodes = tree.get_attribute(attribute) < t
        data += [tree.reconstruct(deleted_nodes)]

    return data, desc

def _show_profiles(profiles, height=None, fname=None, **kwargs):
    assert len(profiles) == 1, 'Show profile only for one attribute at a time.'

    # Set vmin and vmax if not set
    if not 'vmin' in kwargs:
        kwargs['vmin'] = profiles.data.min()
    if not 'vmax' in kwargs:
        kwargs['vmax'] = profiles.data.max()

    if height is not None:
        plt.figure(figsize=_figsize(profiles, height))

    suptitle = '{} - {}'.format(profiles.description['image'], profiles.description['attribute'])

    for i, (im, profile) in enumerate(zip(profiles.data, profiles.description['profiles'])):
        plt.subplot(1, len(profiles.data), i+1)
        plt.imshow(im, **kwargs)
        plt.title(_title(profile))

    plt.tight_layout()
    plt.suptitle(suptitle)

    if fname:
        fname = Path(fname)
        fn = fname.parent / Path(fname.stem + '_{}'.format(profiles.description['attribute']) + fname.suffix)
        plt.savefig(fn)

def show_all_profiles(profiles, attribute=None, image=None, height=None, fname=None, **kwargs):
    """Display profiles with matplotlib.
    
    Parameters
    ----------
    profiles : sap.profiles.Profiles
        The profiles to display.
    attribute : sring, optional
        Name of attribute to display. By default display all the
        attributes contained in profiles.
    image : string, optional
        Name of the image to display. By default display the profiles of
        all images.
    height : scalar, optional, default: None
        Height of the figure in inches. Automatically adjust the size of
        the figure to display correctly the profiles and the title with
        matplot.
    fname : str or PathLike, optional
        If set, the file path to save the figure. The attribute name is
        automatically inserted in the file name.
    
    See Also
    --------
    show_profiles : Display a profiles stack.

    Notes
    -----
    This is a utility function to call recursively `show_profiles`.
    Attribute and image filters are available to filter the profiles to
    display.
    
    """

    # Filter profiles according to attribute if attribute is set
    if attribute:
        profiles = filter(lambda x: x.description['attribute'] == attribute, profiles)
    # Same for image
    if image:
        profiles = filter(lambda x: x.description['image'] == image, profiles)

    for p in profiles:
        show_profiles(p, height, fname, **kwargs)

def differential(profiles):
    """Compute the differential of profiles.

    Parameters
    ----------
    profiles : Profiles
        Attribute profiles or other profiles to process the differential
        on. 
    
    Returns
    -------
    differential : Profiles
        The processed differential profiles.
    
    """
    new_data = []
    new_desc = []

    for p in profiles:
        new_data += [p.data[:-1] - p.data[1:]]
        new_desc += [p.description.copy()]
        d = new_desc[-1]
        d['profiles'] = [{'operation': 'differential', 
                          'profiles': [x, y]} for x, y in zip(d['profiles'][:-1],
                                                              d['profiles'][1:])]
    return Profiles(new_data, new_desc)

def show_profiles(profiles, height=None, fname=None, **kwargs):
    """Display a profiles stack with matplotlib.
    
    Parameters
    ----------
    profiles : Profiles
        The profiles to display. Can be only of length 1.
    height : scalar, optional, default: None
        Height of the figure in inches. Automatically adjust the size of
        the figure to display correctly the profiles and the title with
        matplot.
    fname : str or PathLike, optional
        If set, the file path to save the figure. The attribute name is
        automatically inserted in the file name.

    See Also
    --------
    show_profiles_all : Display several profiles at once.

    """
    _show_profiles(profiles, height, fname, **kwargs)

def _figsize(profiles, height):
    """Compute size of fig given height."""
    shape = profiles.data.shape[1:]
    count = profiles.data.shape[0]
    hw_ratio = shape[1] / shape[0]
    width = height * hw_ratio * count
    return (width, 1.1 * height)
    
def _title(profile):
    """Process a title of a fig."""
    if profile['operation'] == 'differential':
        p1, p2 = profile['profiles']
        return 'differential ({}, {})'.format(_title(p1), _title(p2))
    else:
        return ' '.join([str(x) for x in profile.values()])




#!/usr/bin/env python
# file setup.py
# author Florent Guiotte <florent.guiotte@irisa.fr>
# version 0.0
# date 12 nov. 2019

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='sap',
    version='0.1.0',
    author='Florent Guiotte',
    author_email='florent.guiotte@irisa.fr',
    description='Simple Attribute Profiles',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.inria.fr/fguiotte/sap',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'higra',
        'tqdm',
        'matplotlib',
        'pathlib'
    ],
)



#!/usr/bin/env python
# file Tree.py
# author Florent Guiotte <florent.guiotte@irisa.fr>
# version 0.0
# date 13 nov. 2019
"""
Trees
=====

This submodule contains the component tree classes.

Example
-------

Simple creation of the max-tree of an image, compute the area attributes
of the nodes and reconstruct a filtered image removing nodes with area
less than 100 pixels:

>>>  t = sap.MaxTree(image)
>>>  area = t.get_attribute('area')
>>>  filtered_image = t.reconstruct(area < 100)

"""

import higra as hg
import numpy as np
import inspect

def available_attributes():
    """
    Return a dictionary of available attributes and parameters.

    Returns
    -------
    dict_of_attributes : dict
        The names of available attributes and parameters required.
        The names are keys (str) and the parameters are values (list
        of str) of the dictionary.

    See Also
    --------
    get_attribute : Return the attribute values of the tree nodes.

    Notes
    -----
    The list of available attributes is generated dynamically. It is
    dependent of higra's installed version. For more details, please
    refer to `higra documentation
    <https://higra.readthedocs.io/en/stable/python/tree_attributes.html>`_
    according to the appropriate higra's version.

    Example
    -------
    >>> sap.available_attributes()
    {'area': ['vertex_area=None', 'leaf_graph=None'],
     'compactness': ['area=None', 'contour_length=None', ...],
     ...
     'volume': ['altitudes', 'area=None']}

    """
    params_remove = ['tree']
    dict_of_attributes = {}
    for x in inspect.getmembers(hg):
        if x[0].startswith('attribute_'):
            attribute_name = x[0].replace('attribute_', '')
            attribute_param = \
                [str(x) for x in inspect.signature(x[1]).parameters.values()]
            if attribute_param[0] != 'tree': continue
            attribute_param = \
                list(filter(lambda x: x not in params_remove, attribute_param))
            dict_of_attributes[attribute_name] = attribute_param
    return dict_of_attributes

class Tree:
    """
    Abstract class for tree representations of images.

    Notes
    ----- 
    You should not instantiate class `Tree` directly, use `MaxTree` or
    `MinTree` instead.


    """
    def __init__(self, image, adjacency):
        if self.__class__ == Tree:
            raise TypeError('Do not instantiate directly abstract class Tree.')

        self._adjacency = adjacency
        self._image = image

    def _get_adjacency_graph(self):
        if self._adjacency == 4:
            return hg.get_4_adjacency_graph(self._image.shape)
        elif self._adjacency == 8:
            return hg.get_8_adjacency_graph(self._image.shape)
        else:
            raise NotImplementedError('adjacency of {} is not '
                    'implemented.'.format(self._adjacency))

    def available_attributes(self=None):
        """
        Return a dictionary of available attributes and parameters.

        Returns
        -------
        dict_of_attributes : dict
            The names of available attributes and parameters required.
            The names are keys (str) and the parameters are values (list
            of str) of the dictionary.

        See Also
        --------
        get_attribute : Return the attribute values of the tree nodes.

        Notes
        -----
        The list of available attributes is generated dynamically. It is
        dependent of higra's installed version. For more details, please
        refer to `higra documentation
        <https://higra.readthedocs.io/en/stable/python/tree_attributes.html>`_
        according to the appropriate higra's version.

        Example
        -------
        >>> sap.Tree.available_attributes()
        {'area': ['vertex_area=None', 'leaf_graph=None'],
         'compactness': ['area=None', 'contour_length=None', ...],
         ...
         'volume': ['altitudes', 'area=None']}

        """
        return available_attributes()

    def get_attribute(self, attribute_name, **kwargs):
        """
        Get attribute values of the tree nodes.

        Parameters
        ------
        attribute_name : str
            Name of the attribute (e.g. 'area', 'compactness', ...)

        Returns
        -------
        attribute_values: ndarray
            The values of attribute for each nodes.

        See Also
        --------
        available_attributes : Return the list of available attributes.

        Notes
        -----
        Some attributes require additional parameters. Please refer to
        `available_attributes`. If not stated, some additional
        parameters are automatically deducted. These deducted parameters
        are 'altitudes' and 'vertex_weights'.

        The available attributes depends of higra's installed version.
        For further details Please refer to `higra documentation
        <https://higra.readthedocs.io/en/stable/python/tree_attributes.html>`_
        according to the appropriate higra's version.

        Examples
        --------
        >>> image = np.arange(20 * 50).reshape(20, 50)
        >>> t = sap.MaxTree(image)
        >>> t.get_attribute('area')
        array([   1.,    1.,    1., ...,  998.,  999., 1000.])

        """
        compute = getattr(hg, 'attribute_' + attribute_name)

        if 'altitudes' in inspect.signature(compute).parameters:
            kwargs['altitudes'] = kwargs.get('altitudes', self._alt)

        if 'vertex_weights' in inspect.signature(compute).parameters:
            kwargs['vertex_weights'] = kwargs.get('vertex_weights', self._image)

        return compute(self._tree, **kwargs)

    def reconstruct(self, deleted_nodes=None):
        """
        Return the reconstructed image according to deleted nodes.

        Parameters
        ----------
        deleted_nodes : ndarray or boolean
            Boolean array of node to delete with `len(deleted_nodes) ==
            tree.num_nodes()`.
        
        Returns
        -------
        filtered_image : ndarray
            The reconstructed image.

        """
        if isinstance(deleted_nodes, bool):
            deleted_nodes = np.array((deleted_nodes,) * self.num_nodes())

        return hg.reconstruct_leaf_data(self._tree, self._alt, deleted_nodes)
    
    def num_nodes(self):
        """
        Return the node count of the tree.

        Returns
        -------
        nodes_count : int
            The node count of the tree.

        """
        return self._tree.num_vertices()

class MaxTree(Tree):
    """
    Max tree class, the local maxima values of the image are in leafs.

    Parameters
    ----------
    image : ndarray
        The image to be represented by the tree structure.
    adjacency : int
        The pixel connectivity to use during the tree creation. It
        determines the number of pixels to be taken into account in the
        neighborhood of each pixel. The allowed adjacency are 4 or 8.
        Default is 4.

    Notes
    -----
    Inherits all methods of `Tree` class.

    """
    def __init__(self, image, adjacency=4):
        super().__init__(image, adjacency)
        graph = self._get_adjacency_graph()
        self._tree, self._alt = hg.component_tree_max_tree(graph, image)


class MinTree(Tree):
    """
    Min tree class, the local minima values of the image are in leafs.

    Parameters
    ----------
    image : ndarray
        The image to be represented by the tree structure.
    adjacency : int
        The pixel connectivity to use during the tree creation. It
        determines the number of pixels to be taken into account in the
        neighborhood of each pixel. The allowed adjacency are 4 or 8.
        Default is 4.

    Notes
    -----
    Inherits all methods of `Tree` class.

    """
    def __init__(self, image, adjacency=4):
        super().__init__(image, adjacency)
        graph = self._get_adjacency_graph()
        self._tree, self._alt = hg.component_tree_min_tree(graph, image)
